package statistics

import (
	"context"
	"git.noahlan.cn/northlan/ntools-go/gorm-zero/gormx"
	"github.com/pkg/errors"
	"github.com/zeromicro/go-zero/core/logx"
	"gorm.io/gorm"
	"live-service/app/user_center/model"
	"live-service/app/user_center/rpc/internal/common/integral_manager"
	"live-service/app/user_center/rpc/internal/common/reward_pool"
	"live-service/app/user_center/rpc/internal/svc"
	"live-service/app/user_center/rpc/pb"
	"sort"
)

type StatPvpReportLogic struct {
	ctx    context.Context
	svcCtx *svc.ServiceContext
	logx.Logger
}

func NewStatPvpReportLogic(ctx context.Context, svcCtx *svc.ServiceContext) *StatPvpReportLogic {
	return &StatPvpReportLogic{
		ctx:    ctx,
		svcCtx: svcCtx,
		Logger: logx.WithContext(ctx),
	}
}

func (l *StatPvpReportLogic) StatPvpReport(in *pb.StatPvPReportReq) (*pb.StatPvPReportResp, error) {
	resp := &pb.StatPvPReportResp{}
	// 排序
	sort.SliceStable(in.WinItems, func(i, j int) bool {
		return in.WinItems[i].Position < in.WinItems[j].Position
	})
	sort.SliceStable(in.LostItems, func(i, j int) bool {
		return in.LostItems[i].Position < in.LostItems[j].Position
	})
	// 名将记录
	if len(in.WinItems) > 0 {
		general := in.WinItems[0]
		if general.Uid > 0 {
			if err := l.svcCtx.StatisticsPvpModel.TransactCtx(l.ctx, nil, func(tx *gorm.DB) error {
				if err := l.svcCtx.StatisticsPvpModel.UpdateRecord(l.ctx, tx, general.Uid,
					&model.UpdateRecordProps{General: true}); err != nil {
					if errors.Is(err, gormx.ErrRowsAffectedZero) {
						// insert
						if err = l.svcCtx.StatisticsPvpModel.Insert(l.ctx, tx, &model.StatisticsPvp{
							UserId:       general.Uid,
							GeneralCount: 1,
						}); err != nil {
							return errors.Wrapf(err, ErrInsertErr, err)
						}
					} else {
						return errors.Wrapf(err, ErrUpdateErr, err)
					}
				}
				return nil
			}); err != nil {
				l.Logger.Errorf("PvP名将记录事务执行失败, err:%v", err)
			}
		}
	}
	// 玩家记录
	{
		// 获胜记录
		if err := l.reports(true, in.WinItems); err != nil {
			l.Logger.Errorf("获胜PvP记录失败, err:%v", err)
		}
		// 战败记录
		if err := l.reports(false, in.LostItems); err != nil {
			l.Logger.Errorf("战败PvP记录失败, err:%v", err)
		}
	}
	// 积分记录
	{
		battleReportCfg := l.svcCtx.Config.Integral.BattleReport
		drawReq := make([]reward_pool.DrawRequest, 0, len(in.WinItems))
		for _, item := range in.WinItems {
			drawReq = append(drawReq, reward_pool.DrawRequest{
				UserId:   item.Uid,
				Position: item.Position,
				Ratio:    0.2,
			})
		}
		drawResult := l.svcCtx.RewardPoolManager.Draw(in.BattleId, drawReq)

		winItemResp := make([]*pb.StatPvPReportResp_Item, 0, len(in.WinItems))
		lostItemResp := make([]*pb.StatPvPReportResp_Item, 0, len(in.LostItems))
		integralChangeReq := make([]integral_manager.ChangeIntegralReq, 0, len(in.WinItems)+len(in.LostItems))

		// . 战斗结算
		// . 奖池结算
		for _, item := range in.WinItems {
			// damage (暴兵越多伤害越高)
			// killUnit (暴兵越多击杀越多)
			// deDamage (暴兵越多受伤越多)
			// deKillUnit (暴兵越多死的约多)
			uid, uname, position, damage := item.Uid, item.Uname, item.Position, item.Damage

			battleIntegral := int64(float64(damage) * battleReportCfg.WinRadio)
			tmp := &pb.StatPvPReportResp_Item{
				Uid:                uid,
				Uname:              uname,
				Position:           position,
				ReturnsIntegral:    drawResult.UserReturns[uid],
				RewardPoolIntegral: drawResult.DrawRewards[uid],
				NobilityIntegral:   0, // TODO . 贵族加成分 再议
				BattleIntegral:     battleIntegral,
			}
			if position == 1 {
				tmp.GeneralIntegral = battleReportCfg.GeneralIntegral
			}
			tmp.TotalIntegral = tmp.ReturnsIntegral + tmp.RewardPoolIntegral + tmp.GeneralIntegral + tmp.NobilityIntegral + tmp.BattleIntegral
			winItemResp = append(winItemResp, tmp)
			integralChangeReq = append(integralChangeReq, integral_manager.ChangeIntegralReq{
				BattleId: in.BattleId,
				UserId:   uid,
				Change:   tmp.TotalIntegral,
				Type:     pb.IntegralType_Battle,
			})
		}

		for _, item := range in.LostItems {
			uid, uname, position, damage := item.Uid, item.Uname, item.Position, item.Damage
			battleIntegral := int64(float64(damage) * battleReportCfg.LostRadio)
			tmp := &pb.StatPvPReportResp_Item{
				Uid:                uid,
				Uname:              uname,
				Position:           position,
				ReturnsIntegral:    0,
				RewardPoolIntegral: 0,
				GeneralIntegral:    0,
				NobilityIntegral:   0, // TODO . 贵族加成分 再议
				BattleIntegral:     battleIntegral,
				TotalIntegral:      0,
			}
			tmp.TotalIntegral = tmp.ReturnsIntegral + tmp.RewardPoolIntegral + tmp.GeneralIntegral + tmp.NobilityIntegral + tmp.BattleIntegral
			lostItemResp = append(lostItemResp, tmp)
			integralChangeReq = append(integralChangeReq, integral_manager.ChangeIntegralReq{
				BattleId: in.BattleId,
				UserId:   uid,
				Change:   tmp.TotalIntegral,
				Type:     pb.IntegralType_Battle,
			})
		}
		// 统一结算积分
		_, err := l.svcCtx.IntegralManager.ChangeIntegralBatch(l.ctx, integralChangeReq, false)
		if err != nil {
			l.Logger.Errorf("积分结算失败 %v", err)
		}

		resp.WinItems = winItemResp
		resp.LostItems = lostItemResp
	}
	l.Logger.Info("PvP战报，记录成功...")
	return resp, nil
}

func (l *StatPvpReportLogic) reports(win bool, items []*pb.StatPvPReportReq_Item) error {
	return l.svcCtx.StatisticsPvpModel.TransactCtx(l.ctx, nil, func(tx *gorm.DB) error {
		for _, item := range items {
			if err := l.svcCtx.StatisticsPvpModel.UpdateRecord(l.ctx, tx, item.Uid,
				&model.UpdateRecordProps{
					Win:             win,
					Lost:            !win,
					Damage:          &item.Damage,
					DeDamage:        &item.DeDamage,
					KillUnitCount:   &item.KillUnit,
					DeKillUnitCount: &item.DeKillUnit,
				}); err != nil {
				if errors.Is(err, gormx.ErrRowsAffectedZero) {
					// insert
					var winCount int64
					var lostCount int64
					if win {
						winCount = 1
					} else {
						lostCount = 1
					}
					if err = l.svcCtx.StatisticsPvpModel.Insert(l.ctx, tx, &model.StatisticsPvp{
						UserId:          item.Uid,
						WinCount:        winCount,
						LostCount:       lostCount,
						Damage:          item.Damage,
						DeDamage:        item.DeDamage,
						KillUnitCount:   item.KillUnit,
						DeKillUnitCount: item.DeKillUnit,
					}); err != nil {
						return errors.Wrapf(err, ErrInsertErr, err)
					}
				} else {
					return errors.Wrapf(err, ErrUpdateErr, err)
				}
			}
		}
		return nil
	})
}
