package user

import (
	"context"
	"git.noahlan.cn/northlan/ntools-go/uuid"
	"github.com/pkg/errors"
	"github.com/zeromicro/go-zero/core/logx"
	"gorm.io/gorm"
	"live-service/app/user_center/model"
	"live-service/app/user_center/rpc/internal/svc"
	"live-service/app/user_center/rpc/pb"
	"live-service/common/nerr"
	"time"
)

type RetrievePlatformUserLogic struct {
	ctx    context.Context
	svcCtx *svc.ServiceContext

	logx.Logger
}

func NewRetrievePlatformUserLogic(ctx context.Context, svcCtx *svc.ServiceContext) *RetrievePlatformUserLogic {
	return &RetrievePlatformUserLogic{
		ctx:    ctx,
		svcCtx: svcCtx,
		Logger: logx.WithContext(ctx),
	}
}

// RetrievePlatformUser 查询或创建用户（此时不通过API查询平台用户信息）
func (l *RetrievePlatformUserLogic) RetrievePlatformUser(in *pb.PlatformUserReq) (*pb.PlatformUserResp, error) {
	var dbFullUser *model.FullUser
	var username string

	if err := l.svcCtx.Db.Transaction(func(tx *gorm.DB) error {
		var err error
		if dbFullUser, err = l.svcCtx.UserPlatformModel.FindFullByPlatform(l.ctx, tx, in.Platform, in.PUid); err != nil {
			if !errors.Is(err, model.ErrNotFound) {
				return nerr.NewWithCode(nerr.DBError)
			}
		}
		if dbFullUser != nil {
			if one, err := l.svcCtx.UserModel.FindOne(l.ctx, tx, dbFullUser.UserId); err != nil {
				username = one.Username
			}
			return nil
		}
		// : not found

		// insert
		newId := uuid.NextId()
		if err := l.svcCtx.UserModel.Insert(l.ctx, tx, &model.User{Id: newId}); err != nil {
			return errors.Wrap(nerr.NewWithCode(nerr.DBError), "插入用户数据失败")
		}
		dbPlatformUser := &model.UserPlatform{
			Id:       uuid.NextId(),
			UserId:   newId,
			Platform: in.Platform,
			PUid:     in.PUid,
			PInfo:    "{}",
		}
		if err := l.svcCtx.UserPlatformModel.Insert(l.ctx, tx, dbPlatformUser); err != nil {
			return errors.Wrap(nerr.NewWithCode(nerr.DBError), "插入平台用户数据失败")
		}
		dbFullUser = &model.FullUser{
			UserPlatform: *dbPlatformUser,
			UserNobility: model.UserNobility{
				UserId:        newId,
				NobilityLevel: 0,
			},
			UserIntegral: model.UserIntegral{
				UserId:   newId,
				Integral: 0,
			},
		}
		return nil
	}); err != nil {
		return nil, errors.Wrapf(nerr.NewWithCode(nerr.DBError), "查询或创建用户-事务执行失败, err:%+v", err)
	}
	return &pb.PlatformUserResp{
		User: l.buildPBUser(username, dbFullUser),
	}, nil
}

func (l *RetrievePlatformUserLogic) buildPBUser(username string, dbFullUser *model.FullUser) *pb.User {
	resp := &pb.User{
		Id:            dbFullUser.UserId,
		Username:      username,
		Platform:      dbFullUser.Platform,
		PUid:          dbFullUser.PUid,
		PUname:        dbFullUser.PUname,
		PAvatar:       dbFullUser.PAvatar,
		NobilityLevel: int32(dbFullUser.UserNobility.NobilityLevel),
		Integral:      dbFullUser.UserIntegral.Integral,
	}
	// 贵族过期判断
	if time.Now().After(dbFullUser.UserNobility.EndTime) {
		if !dbFullUser.UserNobility.Forever {
			// 贵族过期，置为0
			resp.NobilityLevel = 0
		}
	}
	return resp
}
